/* Pavel Surynek, 2008 */
/* pavel.surynek@mff.cuni.cz */
/* mit.cpp */

/*
  An implementation of the algorithm for solving sliding box puzzles by
  Kornhauser, Miler, and Spirakis.
 */


#include "slider.h"
#include "mit.h"


void save_fine_solution(FILE *f, const FINE_SOLUTION &fine_solution)
{
	fprintf(f, "Fine solution\n");
	fprintf(f, "Length:%d\n", fine_solution.size());

	for(FINE_SOLUTION::const_iterator move = fine_solution.begin(); move != fine_solution.end(); ++move)
	{
		move->save(f);
	}
}


void load_fine_solution(FILE *f, FINE_SOLUTION &fine_solution)
{
	int length;

	fscanf(f, "Fine solution\n");
	fscanf(f, "Length:%d\n", &length);

	for (int i = 0; i < length; ++i)
	{
		FINE_MOVE move;
		move.load(f);

		fine_solution.push_back(move);
	}
}


void print_fine_solution(const FINE_SOLUTION &fine_solution)
{
	printf("Solution\n");
	for (FINE_SOLUTION::const_iterator move = fine_solution.begin(); move != fine_solution.end(); ++move)
	{
		printf("  %d ---> %d\n", move->m_x, move->m_y);
	}
	printf("Length:%d\n", fine_solution.size());
}


void cycle_1_rotate_L(SLIDER *slider, FINE_SOLUTION &solution)
{
	int i, t;

	t = slider->m_vrtxs[slider->m_cycle1[0]];
	for (i = 1; i < slider->m_C1; ++i)
	{
		slider->m_vrtxs[slider->m_cycle1[i - 1]] = slider->m_vrtxs[slider->m_cycle1[i]];
		solution.push_back(FINE_MOVE(slider->m_cycle1[i], slider->m_cycle1[i - 1]));
	}
	slider->m_vrtxs[slider->m_cycle1[slider->m_C1 - 1]] = t;

	t = slider->m_vrtxs[slider->m_cycle1[slider->m_C1 - 1]];
	slider->m_vrtxs[slider->m_cycle1[slider->m_C1 - 1]] = slider->m_vrtxs[slider->m_cycle1[0]];
	solution.push_back(FINE_MOVE(slider->m_cycle1[0], slider->m_cycle1[slider->m_C1 - 1]));

	slider->m_vrtxs[slider->m_cycle1[0]] = t;
}


void cycle_1_rotate_R(SLIDER *slider, FINE_SOLUTION &solution)
{
	int i, t;

	t = slider->m_vrtxs[slider->m_cycle1[slider->m_C1 - 1]];
	solution.push_back(FINE_MOVE(slider->m_cycle1[slider->m_C1 - 1], slider->m_cycle1[0]));
	
	for (i = slider->m_C1 - 1; i > 0; --i)
	{
		slider->m_vrtxs[slider->m_cycle1[i]] = slider->m_vrtxs[slider->m_cycle1[i - 1]];
		solution.push_back(FINE_MOVE(slider->m_cycle1[i - 1], slider->m_cycle1[i]));
	}
	slider->m_vrtxs[slider->m_cycle1[0]] = t;

	t = slider->m_vrtxs[slider->m_cycle1[1]];
	slider->m_vrtxs[slider->m_cycle1[1]] = slider->m_vrtxs[slider->m_cycle1[0]];
	slider->m_vrtxs[slider->m_cycle1[0]] = t;
}


void cycle_2_rotate_L(SLIDER *slider, FINE_SOLUTION &solution)
{
	int i, t;

	t = slider->m_vrtxs[slider->m_cycle2[0]];

	for (i = 1; i < slider->m_C2; ++i)
	{
		slider->m_vrtxs[slider->m_cycle2[i - 1]] = slider->m_vrtxs[slider->m_cycle2[i]];
		solution.push_back(FINE_MOVE(slider->m_cycle2[i], slider->m_cycle2[i - 1]));
	}
	slider->m_vrtxs[slider->m_cycle2[slider->m_C2 - 1]] = t;

	t = slider->m_vrtxs[slider->m_cycle2[slider->m_C2 - 1]];
	slider->m_vrtxs[slider->m_cycle2[slider->m_C2 - 1]] = slider->m_vrtxs[slider->m_cycle2[0]];
	solution.push_back(FINE_MOVE(slider->m_cycle2[0], slider->m_cycle2[slider->m_C2 - 1]));

	slider->m_vrtxs[slider->m_cycle2[0]] = t;
}


void cycle_2_rotate_R(SLIDER *slider, FINE_SOLUTION &solution)
{
	int i, t;

	t = slider->m_vrtxs[slider->m_cycle2[slider->m_C2 - 1]];
	solution.push_back(FINE_MOVE(slider->m_cycle2[slider->m_C2 - 1], slider->m_cycle2[0]));
	
	for (i = slider->m_C2 - 1; i > 0; --i)
	{
		slider->m_vrtxs[slider->m_cycle2[i]] = slider->m_vrtxs[slider->m_cycle2[i - 1]];
		solution.push_back(FINE_MOVE(slider->m_cycle2[i - 1], slider->m_cycle2[i]));
	}
	slider->m_vrtxs[slider->m_cycle2[0]] = t;

	t = slider->m_vrtxs[slider->m_cycle2[1]];
	slider->m_vrtxs[slider->m_cycle2[1]] = slider->m_vrtxs[slider->m_cycle2[0]];
	slider->m_vrtxs[slider->m_cycle2[0]] = t;
}


void cycle_3_rotate_L(SLIDER *slider, FINE_SOLUTION &solution)
{
	int i, t;

	t = slider->m_vrtxs[slider->m_cycle3[0]];

	for (i = 1; i < slider->m_C3; ++i)
	{
		slider->m_vrtxs[slider->m_cycle3[i - 1]] = slider->m_vrtxs[slider->m_cycle3[i]];
		solution.push_back(FINE_MOVE(slider->m_cycle3[i], slider->m_cycle3[i - 1]));
	}
	slider->m_vrtxs[slider->m_cycle3[slider->m_C3 - 1]] = t;


	t = slider->m_vrtxs[slider->m_cycle3[slider->m_C3 - 1]];
	slider->m_vrtxs[slider->m_cycle3[slider->m_C3 - 1]] = slider->m_vrtxs[slider->m_cycle3[0]];
	solution.push_back(FINE_MOVE(slider->m_cycle3[0], slider->m_cycle3[slider->m_C3 - 1]));

	slider->m_vrtxs[slider->m_cycle3[0]] = t;
}


void cycle_3_rotate_R(SLIDER *slider, FINE_SOLUTION &solution)
{
	int i, t;

	t = slider->m_vrtxs[slider->m_cycle3[slider->m_C3 - 1]];
	solution.push_back(FINE_MOVE(slider->m_cycle3[slider->m_C3 - 1], slider->m_cycle3[0]));
	
	for (i = slider->m_C3 - 1; i > 0; --i)
	{
		slider->m_vrtxs[slider->m_cycle3[i]] = slider->m_vrtxs[slider->m_cycle3[i - 1]];
		solution.push_back(FINE_MOVE(slider->m_cycle3[i - 1], slider->m_cycle3[i]));
	}
	slider->m_vrtxs[slider->m_cycle3[0]] = t;

	t = slider->m_vrtxs[slider->m_cycle3[1]];
	slider->m_vrtxs[slider->m_cycle3[1]] = slider->m_vrtxs[slider->m_cycle3[0]];
	slider->m_vrtxs[slider->m_cycle3[0]] = t;
}


int find_robot(SLIDER *slider, int r)
{
	for (int i = 0; i < slider->m_V; ++i)
	{
		if (slider->m_vrtxs[i] == r)
		{
			return i;
		}
	}

	return -1;
}


int find_robot_in_cycle_1(SLIDER *slider, int r)
{
	for (int i = 0; i < slider->m_C1; ++i)
	{
		if (slider->m_vrtxs[slider->m_cycle1[i]] == r)
		{
			return i;
		}
	}

	return -1;
}


int find_robot_in_cycle_2(SLIDER *slider, int r)
{
	for (int i = 0; i < slider->m_C2; ++i)
	{
		if (slider->m_vrtxs[slider->m_cycle2[i]] == r)
		{
			return i;
		}
	}

	return -1;
}


int find_robot_in_cycle_3(SLIDER *slider, int r)
{
	for (int i = 0; i < slider->m_C3; ++i)
	{
		if (slider->m_vrtxs[slider->m_cycle3[i]] == r)
		{
			return i;
		}
	}

	return -1;
}


void move_4_transitively_cycle_1(SLIDER *slider, int *robots, int *positions, FINE_SOLUTION &solution)
{
	move_4_transitively_cycle_1_half(slider, robots, solution);

	SLIDER copy;
	copy_slider(slider, &copy);

	int p_robots[4];

	for (int i = 0; i < 4; ++i)
	{
		assert(slider->m_vrtxs[positions[i]] > 0);
		p_robots[i] = slider->m_vrtxs[positions[i]];
	}

	FINE_SOLUTION solution_half;
	move_4_transitively_cycle_1_half(&copy, p_robots, solution_half);

	for (FINE_SOLUTION::iterator move = solution_half.begin(); move != solution_half.end(); ++move)
	{
		move->reverse();
	}
	solution.insert(solution.end(), solution_half.rbegin(), solution_half.rend());

	destroy_slider(&copy);
}


void move_4_transitively_cycle_1_half(SLIDER *slider, int *robots, FINE_SOLUTION &solution)
{
	for (int i = 0; i < 4; ++i)
	{
		int r = robots[i];
		int c1 = find_robot_in_cycle_1(slider, r);
		int c2 = find_robot_in_cycle_2(slider, r);

		if (c1 >= 0)
		{
			if (c2 >= 0)
			{
				if (c2 <= slider->m_C2 / 2)
				{
					while (slider->m_vrtxs[slider->m_cycle2[1]] != r)
					{
						cycle_2_rotate_L(slider, solution);
					}
				}
				else
				{
					while (slider->m_vrtxs[slider->m_cycle2[1]] != r)
					{
						cycle_2_rotate_R(slider, solution);
					}
				}
				cycle_1_rotate_R(slider, solution);
			}
			else
			{
				if (i == 0)
				{
					if (c1 <= slider->m_C1 / 2)
					{
						while (slider->m_vrtxs[slider->m_cycle1[1]] != r)
						{
							cycle_1_rotate_L(slider, solution);
						}
					}
					else
					{
						while (slider->m_vrtxs[slider->m_cycle1[1]] != r)
						{
							cycle_1_rotate_R(slider, solution);
						}
					}
				}
				else
				{
					int c = 0;
					while (slider->m_vrtxs[slider->m_Y] != r)
					{
						cycle_1_rotate_L(slider, solution);
						++c;
					}
					cycle_2_rotate_R(slider, solution);
					
					while (c > 0)
					{
						cycle_1_rotate_R(slider, solution);
						--c;
					}
					
					c2 = find_robot_in_cycle_2(slider, r);

					if (c2 <= slider->m_C2 / 2)
					{
						while (slider->m_vrtxs[slider->m_cycle1[slider->m_C1 - 1]] != r)
						{
							cycle_2_rotate_L(slider, solution);
						}
					}
					else
					{
						while (slider->m_vrtxs[slider->m_cycle1[slider->m_C1 - 1]] != r)
						{
							cycle_2_rotate_R(slider, solution);
						}
					}
					cycle_1_rotate_R(slider, solution);
				}
			}
		}
		else
		{
			if (c2 >= 0)
			{
				if (c2 <= slider->m_C2 / 2)
				{
					while (slider->m_vrtxs[slider->m_cycle2[1]] != r)
					{
						cycle_2_rotate_L(slider, solution);
					}
				}
				else
				{
					while (slider->m_vrtxs[slider->m_cycle2[1]] != r)
					{
						cycle_2_rotate_R(slider, solution);
					}
				}
				cycle_1_rotate_R(slider, solution);
			}
			else
			{
				assert(false);  // It cannot happen that the robot is not in cycle 1 and cycle 2
				return;
			}
		}
	}
}


void execute_fine_solution(SLIDER *slider, const FINE_SOLUTION &solution)
{
	for (FINE_SOLUTION::const_iterator move = solution.begin(); move != solution.end(); ++move)
	{
		if (slider->m_vrtxs[move->m_x] > 0 && slider->m_vrtxs[move->m_y] <= 0)
		{
			int t = slider->m_vrtxs[move->m_y];
			slider->m_vrtxs[move->m_y] = slider->m_vrtxs[move->m_x];
			slider->m_vrtxs[move->m_x] = t;
		}
		else
		{
			print_slider(slider);
			printf("Invalid move: %d,%d\n", move->m_x, move->m_y);
			assert(false);
		}
	}
}


void generate_3_cycle(SLIDER *slider, int x, int y, int z, FINE_SOLUTION &solution)
{
	if (slider->m_C1 - slider->m_Y > 1)
	{
		generate_3_cycle_CASE_B(slider, x, y, z, solution);
	}
	else
	{
		generate_3_cycle_CASE_A(slider, x, y, z, solution);
	}
}


void generate_3_cycle_CASE_A(SLIDER *slider, int x, int y, int z, FINE_SOLUTION &solution)
{
	SLIDER copy;
	copy_slider(slider, &copy);

	FINE_SOLUTION first_phase, second_phase;

	int dummy;

	for (int i = 1; i < slider->m_V; ++i)
	{
		if (i != x && i != y && i != z)
		{
			dummy = i;
			break;
		}
	}

	int robots[] = {x, y, z, dummy};
	int positions[] = {1, slider->m_Y, slider->m_C1, 2};

	move_4_transitively_cycle_1(&copy, robots, positions, first_phase);

	cycle_1_rotate_L(&copy, second_phase);
	cycle_2_rotate_L(&copy, second_phase);
	cycle_1_rotate_R(&copy, second_phase);
	cycle_2_rotate_R(&copy, second_phase);

	solution.insert(solution.end(), first_phase.begin(), first_phase.end());
	solution.insert(solution.end(), second_phase.begin(), second_phase.end());

	for (FINE_SOLUTION::iterator move = first_phase.begin(); move != first_phase.end(); ++move)
	{
		move->reverse();
	}
	solution.insert(solution.end(), first_phase.rbegin(), first_phase.rend());	

	destroy_slider(&copy);
}


void generate_3_cycle_CASE_B(SLIDER *slider, int x, int y, int z, FINE_SOLUTION &solution)
{
	FINE_SOLUTION first_phase, second_phase, third_phase, fourth_phase;

	SLIDER copy1, copy2;
	copy_slider(slider, &copy1);
	copy_slider(slider, &copy2);

	int a, b;

	for (int i = 1; i < slider->m_V; ++i)
	{
		if (i != x && i != y && i != z)
		{
			a = i;
			break;
		}
	}

	for (int i = 1; i < slider->m_V; ++i)
	{
		if (i != x && i != y && i != z && i != a)
		{
			b = i;
			break;
		}
	}

	int robots_1[] = {a, b, x, z};
	int positions_1[] = {1, slider->m_C1 - 1, slider->m_Y, slider->m_C1};

	SLIDER copy3;
	copy_slider(&copy1, &copy3);

	move_4_transitively_cycle_1(&copy1, robots_1, positions_1, first_phase);
	execute_fine_solution(&copy3, first_phase);

	cycle_1_rotate_L(&copy1, second_phase);
	cycle_2_rotate_L(&copy1, second_phase);
	cycle_1_rotate_R(&copy1, second_phase);
	cycle_2_rotate_R(&copy1, second_phase);

	solution.insert(solution.end(), first_phase.begin(), first_phase.end());
	solution.insert(solution.end(), second_phase.begin(), second_phase.end());

	for (FINE_SOLUTION::iterator move = first_phase.begin(); move != first_phase.end(); ++move)
	{
		move->reverse();
	}
	solution.insert(solution.end(), first_phase.rbegin(), first_phase.rend());	

	int robots_2[] = {a, b, z, y};
//	int positions_2[] = {1, 7, 5, 8};
	int positions_2[] = {1, slider->m_C1 - 1, slider->m_Y, slider->m_C1};

	move_4_transitively_cycle_1(&copy2, robots_2, positions_2, third_phase);

	cycle_1_rotate_L(&copy2, fourth_phase);
	cycle_2_rotate_L(&copy2, fourth_phase);
	cycle_1_rotate_R(&copy2, fourth_phase);
	cycle_2_rotate_R(&copy2, fourth_phase);

	solution.insert(solution.end(), third_phase.begin(), third_phase.end());
	solution.insert(solution.end(), fourth_phase.begin(), fourth_phase.end());

	for (FINE_SOLUTION::iterator move = third_phase.begin(); move != third_phase.end(); ++move)
	{
		move->reverse();
	}
	solution.insert(solution.end(), third_phase.rbegin(), third_phase.rend());	

	destroy_slider(&copy1);
	destroy_slider(&copy2);
}


int calculate_transpositions(SLIDER *slider)
{
	int trans = 0;

	for (int i = 1; i < slider->m_V - 1; ++i)
	{
		for (int j = i + 1; j < slider->m_V; ++j)
		{
			if (slider->m_vrtxs[i] > slider->m_vrtxs[j])
			{
				++trans;
			}
		}
	}

	return trans;
}


int calculate_goal_transpositions(SLIDER *slider)
{
	int trans = 0;

	for (int i = 1; i < slider->m_V - 1; ++i)
	{
		for (int j = i + 1; j < slider->m_V; ++j)
		{
			if (slider->m_goal[i] > slider->m_goal[j])
			{
				++trans;
			}
		}
	}

	return trans;
}


bool solve_slider_mit(SLIDER *slider, FINE_SOLUTION &solution)
{
	if (slider->m_Y <= 3)
	{
//		printf("Too small first handle. Unable to solve. Terminating...\n");
		return false;
	}

	std::vector<int> forbid;

	SLIDER copy;
	copy_slider(slider, &copy);

	FINE_SOLUTION rotation_solution;

	int trans = calculate_goal_transpositions(&copy);

	if (trans % 2 == 1)  // odd permutation, we need to find odd cycle
	{
		if (copy.m_C1 % 2 == 1)
		{
			cycle_1_rotate_L(&copy, rotation_solution);
		}
		else
		{
			if (copy.m_C2 % 2 == 1)
			{
				cycle_2_rotate_L(&copy, rotation_solution);
			}
			else
			{
				if (copy.m_C3 % 2 == 1)
				{
					cycle_3_rotate_L(&copy, rotation_solution);
				}
				else
				{
					printf("Unable to solve, cannot obtain odd permutation using 3-cycles. Terminating...\n");
					assert(false);
				}
			}
		}
	}
	solution.insert(solution.end(), rotation_solution.begin(), rotation_solution.end());

	forbid.push_back(copy.m_goal[1]);

	for (int i = 1; i < copy.m_V; ++i)
	{
		forbid.push_back(copy.m_vrtxs[i]);
		forbid.push_back(copy.m_goal[i]);

		if (copy.m_vrtxs[i] != copy.m_goal[i])
		{
			FINE_SOLUTION phase_solution;

			int z = -1;
			
			for (int j = 1; j < copy.m_V; ++j)
			{
				bool found = false;
				for (std::vector<int>::const_iterator f = forbid.begin(); f != forbid.end(); ++f)
				{
					if (*f == j)
					{
						found = true;
						break;
					}
				}
				if (!found)
				{
					z = j;
					break;
				}
			}

			if (z == -1)
			{
				break;
			}

			generate_3_cycle(&copy, copy.m_goal[i], copy.m_vrtxs[i], z, phase_solution);
			solution.insert(solution.end(), phase_solution.begin(), phase_solution.end());
			
			execute_fine_solution(&copy, phase_solution);
		}
	}

	destroy_slider(&copy);

	return true;
}
