/* Pavel Surynek, 2008 */
/* pavel.surynek@mff.cuni.cz */
/* mit_main.cpp */

/*
  An implementation of the algorithm for solving sliding box puzzles by
  Kornhauser, Miler, and Spirakis.
 */

#include "slider.h"
#include "mit.h"
#include "tbox.h"


SLIDER slider1, slider2, slider3;
SOLUTION_DATABASE database;


int main(int UNUSED(argc), char **UNUSED(argv))
{
	clock_t begin, end;

//	generate_solution_database("solution_database_B.txt");
//	generate_solution_3_database("solution_database_C.txt");
//	init_slider(0, 7, 5, 10, &slider1);

	database.load("solution_database_B.txt");

	int c1_b, c2_a, c2_b;

	for (c1_b = 3; c1_b < 16; ++c1_b)
	{
		for (c2_a = 1; c2_a <= c1_b; ++c2_a)
		{
			for (c2_b = c1_b + 1; c1_b - c2_a + c2_b - c1_b + 1 <= c1_b; ++c2_b)
			{
				init_slider(0, c1_b, c2_a, c2_b, &slider1);

				int oddness = slider1.m_C1 % 2 || slider1.m_C2 % 2 || slider1.m_C3 % 2;

				clock_t begin_tbox, end_tbox;
				clock_t begin_mit, end_mit;
				
				if (oddness && (c1_b != 5 || c2_a != 3 || c2_b != 6))
				{
					printf("A:%d\n", slider1.m_A);
					printf("B:%d\n", slider1.m_B);
					printf("C:%d\n", slider1.m_C);

					begin = clock();

					int total_length = 0;
					int total_count = 0;
		
					for (int l = 0; l < 1000; ++l)
					{
						FINE_SOLUTION solution, mit_solution;
						
						copy_slider(&slider1, &slider2);
						copy_slider(&slider1, &slider3);

						for (int i = 0; i < slider2.m_V * slider2.m_V; ++i)
						{
							int pos1 = rand() % (slider2.m_V - 1) + 1;
							int pos2 = (pos1 + 1) % (slider2.m_V - 1) + 1;
							
							swap_goal(pos1, pos2, &slider2);
							swap_goal(pos1, pos2, &slider3);
						}
						
						bool solved = solve_slider_tbox(&slider2, database, solution, 1.0);

						if (solved)
						{
							total_length += solution.size();
							++total_count;
						}

/*
						bool solved = solve_slider_mit(&slider3, mit_solution);

						if (solved)
						{
							total_length += mit_solution.size();
							++total_count;
						}
*/

						destroy_slider(&slider2);
						destroy_slider(&slider3);
					}

					printf("Total: %d,%d\n", total_length, total_count);
					printf("Average solution length: %.3f\n", (double)total_length / (double)total_count);
					end = clock();
					printf("Time:%.3f\n", (end - begin) / (double)CLOCKS_PER_SEC);
					printf("Average time:%f\n", ((end - begin) / (double)CLOCKS_PER_SEC) / (double)total_count);
				}
			}
		}
		++c1_b;
	}

	return 0;
}
